cat
	Class;
	Instance Class;
	Subtype (sub, super : Class);

	Constraint;
	ConstraintBody;
	ConstraintHideContext;
	ConstraintHCInvariant;

	VarSelf (c : Class);
	VarResult (c : Class);

	InstIter (iter, exp : Class);
	SentIter (iter : Class);
	CollList Class;

	Sent;
	
	LetExp Class;
	LetExpSent;

fun	
	self : (c:Class) -> VarSelf c -> Instance c;
	result : (c:Class) -> VarResult c -> Instance c;
	
	atPre : (c:Class) -> Instance c -> Instance c;
	atPreSent : Sent -> Sent;
	
	emptyC : ConstraintBody;
	invC : Sent -> ConstraintBody -> ConstraintBody;
	preC : Sent -> ConstraintBody -> ConstraintBody;
	postC : Sent -> ConstraintBody -> ConstraintBody;
	invHC : Sent -> ConstraintHCInvariant;
 
 	subTypeRefl : (c : Class) -> Subtype c c;
 	subTypeTrans : (a, b, c : Class) -> Subtype a b -> Subtype b c -> Subtype a c;
 	coerce : (sub,super:Class) -> Subtype sub super -> Instance sub -> Instance super;
 
	bool2sent : Instance BoolC -> Sent;
	sent2bool : Sent -> Instance BoolC;
	

--
-- Let expressions
-- 	
 	letExp2Inst : (c:Class) -> LetExp c -> Instance c;
 	letSingleAttr : (c1,c2:Class) -> Instance c1 -> (Instance c1 -> Instance c2) -> LetExp c2;
 	letConsAttr : (c1,c2:Class) -> Instance c1 -> (Instance c1 -> LetExp c2) -> LetExp c2;
	letSingleOp1arg : (c1,c2,c3:Class) -> (Instance c1 -> Instance c2) -> 
		((Instance c1 -> Instance c2) -> LetExp c3);

-- 	argSingle : (c1,c2:Class) -> (Instance c1 -> Instance c2) -> LetOpBinding c2;
-- 	argCons : (c1,c2:Class) -> (Instance c1 -> LetOpBinding c2) -> LetOpbinding c2;
 	
 	letExp2Sent :  LetExpSent -> Sent;
 	letSingleAttrS : (c1:Class) -> Instance c1 -> (Instance c1 -> Sent) -> LetExpSent;
 	letConsAttrS : (c1:Class) -> Instance c1 -> (Instance c1 -> LetExpSent) -> LetExpSent;
 	
	


--------------------


----------------------------------------------
-- OCL Library classes
----------------------------------------------

--
-- OclType. Would it be better to just have Class? Apparently not.
--
fun OclType : Class; -- could try to have OclType : Class -> Class ???

-- this is not part of OCL:
fun class2oclType : Class -> Instance OclType;

fun otName : Instance OclType -> Instance StringC;
    otAttributes : Instance OclType -> Instance (Set StringC);
    otOperations : Instance OclType -> Instance (Set StringC);
    otSupertypes : Instance OclType -> Instance (Set OclType);
    otAllSupertypes : Instance OclType -> Instance (Set OclType);
    otAllInstancesStrict : (c:Class) -> Instance OclType -> Instance (Set c); -- Note extra Class argument!
    otAllInstancesClass : (c:Class) -> Instance (Set c); -- for convenience, not in OCL
    otAllInstancesExpr : (c:Class) -> Instance c -> Instance (Set c); -- for convenience, not in OCL
    -- (compos. of allinst and evalType)

--
-- OclAny
--

fun OclAny : Class;
    conforms2OclAny : (c:Class) -> Subtype c OclAny;

-- might be better to define these for all classes (the class being a parameter)?
    anyEq : (a,b : Instance OclAny) -> Sent;
    anyNeq : (a,b : Instance OclAny) -> Sent;
    anyOclIsKindOf : Instance OclAny -> Instance OclType -> Sent;
    anyOclIsTypeOf : Instance OclAny -> Instance OclType -> Sent;    
    anyOclAsTypeStrict : Instance OclAny -> (c:Class) -> Instance OclType -> Instance c; -- N.B.
    anyOclAsTypeClass : Instance OclAny -> (c:Class) -> Instance c; -- for convenience, not in OCL
    anyOclAsTypeExpr : Instance OclAny -> (c:Class) -> Instance c -> Instance c; -- for convenience, not in OCL
    anyOclInState : Instance OclAny -> Instance OclState -> Sent;
    anyOclIsNew : Instance OclAny -> Sent;



--
-- OclState
--
fun OclState : Class;




--
-- OclExpression
--
fun OclExpression : Class;

-- not in OCL:
fun inst2OclExpr : (c:Class) -> Instance c -> Instance OclExpression;

fun oeEvaluationType : Instance OclExpression -> Instance OclType;



--
-- Bool
--   (it might be good to have duplicates of the connectives also on Bool level)
fun	BoolC : Class;
	true : Instance BoolC; false : Instance BoolC;
	boolEq : Instance BoolC -> Instance BoolC -> Sent;
	orBool : Instance BoolC -> Instance BoolC -> Instance BoolC;
	xorBool : Instance BoolC -> Instance BoolC -> Instance BoolC;
	andBool : (p,q : Instance BoolC) -> Instance BoolC;
	notBool : Instance BoolC -> Instance BoolC;
	impliesBool : (p,q : Instance BoolC) -> Instance BoolC;
	-- no ifThenElseBool. 

--
-- Sent 
--
	sentEq : Sent -> Sent -> Sent;
	orS : Sent -> Sent -> Sent;
	xorS : Sent -> Sent -> Sent;
	andS : (p,q : Sent) -> Sent;
	notS : Sent -> Sent;
	impliesS : (p,q : Sent) -> Sent;
	ifThenElse : (c:Class) -> Sent -> Instance c -> Instance c -> Instance c;
	ifThenElseS : Sent -> Sent -> Sent -> Sent;
	
--
-- Integer
--
fun	Integer : Class;
	intEq : Instance Integer -> Instance Integer -> Sent;
	intAdd : Instance Integer -> Instance Integer -> Instance Integer;
	intSub : Instance Integer -> Instance Integer -> Instance Integer;
	intUnaryMinus : Instance Integer -> Instance Integer;
	intTimes : Instance Integer -> Instance Integer -> Instance Integer;
	intDiv2Real : Instance Integer -> Instance Integer -> Instance Real;
	intAbs : Instance Integer -> Instance Integer;
	intDiv2Int : Instance Integer -> Instance Integer -> Instance Integer;
	intMod : Instance Integer -> Instance Integer -> Instance Integer;
	intMax : Instance Integer -> Instance Integer -> Instance Integer;
	intMin : Instance Integer -> Instance Integer -> Instance Integer;
	-- below are comparison operators supplied for convenience, they are not in OCL:
	intLT : Instance Integer -> Instance Integer -> Sent;
	intGT : Instance Integer -> Instance Integer -> Sent;
	intLTE : Instance Integer -> Instance Integer -> Sent;
	intGTE : Instance Integer -> Instance Integer -> Sent;
	
	
	Zero : Instance Integer; One : Instance Integer; Two : Instance Integer;
	int : Int -> Instance Integer;

--
-- Real
--
	
fun	Real : Class;
	realEq : Instance Real -> Instance Real -> Sent;
	realNeq : Instance Real -> Instance Real -> Sent;
	realAdd : (a,b : Instance Real) -> Instance Real;
	realSub : (a,b : Instance Real) -> Instance Real;
	realTimes : (a,b : Instance Real) -> Instance Real;
	realDiv : (a,b : Instance Real) -> Instance Real;
	realUnaryMinus : Instance Real -> Instance Real;
	realAbs : Instance Real -> Instance Real;
	realFloor : Instance Real -> Instance Integer;
	realRound : Instance Real -> Instance Integer;
	realMax : (a,b : Instance Real) -> Instance Real;
	realMin : (a,b : Instance Real) -> Instance Real;
	realLT : (a,b : Instance Real) -> Sent;
	realGT : (a,b : Instance Real) -> Sent;
	realLTE : (a,b : Instance Real) -> Sent;
	realGTE : (a,b : Instance Real) -> Sent;


--
-- Strings
--
	
fun	StringC : Class;
	stringEq : (a,b : Instance StringC) -> Sent;
	stringLiteral : String -> Instance StringC;
	stringSize : Instance StringC -> Instance Integer;
	stringConcat : (a,b : Instance StringC) -> Instance StringC;
	stringToUpper : Instance StringC -> Instance StringC;
	stringToLower : Instance StringC -> Instance StringC;
	stringSubstring : Instance StringC -> (a,b : Instance Integer) -> Instance StringC;

--
-- Iterators
--
fun
	instIterSingle : (iter, e : Class) -> (Instance iter -> Instance e) -> InstIter iter e;
	instIterMany : (iter, e : Class) -> (Instance iter -> InstIter iter e) -> InstIter iter e;
	sentIterSingle : (iter : Class) -> (Instance iter -> Sent) -> SentIter iter;
	sentIterMany : (iter : Class) -> (Instance iter -> SentIter iter) -> SentIter iter;
--
-- Collection
--
fun	Collection : Class -> Class;
	size : (c:Class) -> (o : Instance (Collection c)) -> Instance Integer;
	includes : (c,d:Class) -> Instance (Collection c) -> Instance d ->  Sent; -- no use of OclAny
	excludes : (c,d:Class) -> Instance (Collection c) -> Instance d -> Sent;
	count : (c:Class) -> Instance (Collection c) -> Instance c -> Instance Integer;
	includesAll : (c:Class) -> (a,b : Instance (Collection c)) -> Sent;
	excludesAll : (c:Class) -> (a,b : Instance (Collection c)) -> Sent;
	isEmpty : (c:Class) -> Instance (Collection c) -> Sent;
	notEmpty : (c:Class) -> Instance (Collection c) -> Sent;
	sum : (c:Class) -> Instance (Collection c) -> Instance c; -- note diff from OCL

	exists : (c:Class) -> Instance (Collection c) -> SentIter c -> Sent;
	forAll : (c:Class) -> Instance (Collection c) -> SentIter c -> Sent;
	isUnique : (c,d : Class) -> Instance (Collection c) -> (InstIter c d) -> 
		Sent;
	sortedBy : (c,d : Class) -> Instance (Collection c) -> (InstIter c d) -> 
		Instance (Sequence c);  -- note diff OCL
	iterate : (c, a: Class) -> (o: Instance (Collection c)) ->
		(init : Instance a) -> (Instance c -> Instance a -> Instance a) -> Instance a;
	any : (c:Class) -> Instance (Collection c) -> SentIter c -> Instance c;
	one : (c:Class) -> Instance (Collection c) -> SentIter c -> Sent;

--
-- sets
--
	Set : Class -> Class;
	unionSS : (c:Class) -> Instance (Set c) -> Instance (Set c) -> Instance (Set c);
	unionSB : (c:Class) -> Instance (Set c) -> Instance (Bag c) -> Instance (Bag c);
	setEq : (c:Class) -> Instance (Set c) -> Instance (Set c) -> Sent;
	intersectionSS : (c:Class) -> (a,b : Instance (Set c)) -> Instance (Set c);
	intersectionSB : (c:Class) -> Instance (Set c) -> Instance (Bag c) -> Instance (Set c);
	subS : (c:Class) -> (a,b : Instance (Set c)) -> Instance (Set c);
	includingS : (c:Class) -> Instance (Set c) -> Instance c -> Instance (Set c);
	excludingS : (c:Class) -> Instance (Set c) -> Instance c -> Instance (Set c);
	symmetricDiff : (c:Class) -> (a,b : Instance (Set c)) -> Instance (Set c);
	selectSet : (c:Class) -> Instance (Set c) -> SentIter c -> Instance (Set c);
	rejectSet : (c:Class) -> Instance (Set c) -> SentIter c -> Instance (Set c);
	collectSet : (c,d:Class) -> Instance (Set c) -> InstIter c d -> Instance (Bag d);
	countSet : (c:Class) -> Instance (Set c) -> Instance c -> Instance Integer;
	setAsSequence : (c:Class) -> Instance (Set c) -> Instance (Sequence c);
	setAsBag : (c:Class) -> Instance (Set c) -> Instance (Bag c);

--	
-- bags
--
	Bag : Class -> Class;
	bagEq : (c:Class) -> (a,b : Instance (Bag c)) -> Sent;
	unionBB : (c:Class) -> (a,b : Instance (Bag c)) -> Instance (Bag c);
	unionBS : (c:Class) -> Instance (Bag c) -> Instance (Set c) -> Instance (Bag c);
	intersectionBB : (c:Class) -> (a,b : Instance (Bag c)) -> Instance (Bag c); 
	intersectionBS : (c:Class) -> Instance (Bag c) -> Instance (Set c) -> Instance (Set c);
	includingB : (c:Class) -> Instance (Bag c) -> Instance c -> Instance (Bag c);
	excludingB : (c:Class) -> Instance (Bag c) -> Instance c -> Instance (Bag c);
	selectBag : (c:Class) -> Instance (Bag c) -> SentIter c -> Instance (Bag c);
	rejectBag : (c:Class) -> Instance (Bag c) -> SentIter c -> Instance (Bag c);
	collectBag : (c,d:Class) -> Instance (Bag c) -> InstIter c d -> Instance (Bag d);
	countBag : (c:Class) -> Instance (Bag c) -> Instance c -> Instance Integer;
	bagAsSequence : (c:Class) -> Instance (Bag c) -> Instance (Sequence c);
	bagAsSet : (c:Class) -> Instance (Bag c) -> Instance (Set c);

	-- sequences	
	Sequence : Class -> Class;
	sequenceEq : (c:Class) -> (a,b : Instance (Sequence c)) -> Sent;
	unionSeq : (c:Class) -> (a,b : Instance (Sequence c)) -> Instance (Sequence c);
	append : (c:Class) -> Instance (Sequence c) -> Instance c ->
		Instance (Sequence c);
	prepend : (c:Class) -> Instance (Sequence c) -> Instance c ->
		Instance (Sequence c);
	subSequence : (c:Class) -> Instance (Sequence c) -> 
		(a,b : Instance Integer) -> Instance (Sequence c);
	at : (c:Class) -> Instance (Sequence c) -> Instance Integer -> Instance c;
	first : (c:Class) -> (o : Instance (Sequence c)) -> Instance c;
	last : (c:Class) -> (o: Instance (Sequence c)) -> Instance c;
	includingSeq : (c:Class) -> Instance (Sequence c) -> Instance c -> Instance (Sequence c);
	excludingSeq : (c:Class) -> Instance (Sequence c) -> Instance c -> Instance (Sequence c);
	selectSeq : (c:Class) -> Instance (Sequence c) -> SentIter c  -> Instance (Sequence c);
	rejectSeq : (c:Class) -> Instance (Sequence c) -> SentIter c  -> Instance (Sequence c);
	collectSeq : (c,d : Class) -> Instance (Sequence c) -> InstIter c d -> Instance (Sequence d);
	-- no particular iterate for Sequences...
	seqAsBag : (c:Class) -> Instance (Sequence c) -> Instance (Bag c);
	seqAsSet : (c:Class) -> Instance (Sequence c) -> Instance (Set c);

{-	nilColl : (c:Class) -> CollList c; 
	consColl : (c:Class) -> Instance c -> CollList c -> CollList c;
	listAsSeq : (c:Class) -> CollList c -> Instance (Sequence c);
	-}	


