-- language-independent prelude facilities

oper

-- to construct records and tables
  SS  : Type = {s : Str} ;
  ss  : Str -> SS = \s -> {s = s} ;
  ss2 : (_,_ : Str) -> SS = \x,y -> ss (x ++ y) ;
  ss3 : (_,_ ,_: Str) -> SS = \x,y,z -> ss (x ++ y ++ z) ;

  cc2 : (_,_ : SS) -> SS = \x,y -> ss (x.s ++ y.s) ;

  SS1 : Type -> Type = \P -> {s : P => Str} ;
  ss1 : (A : Type) -> Str -> SS1 A = \_,s -> {s = table {_ => s}} ;

  SP1 : Type -> Type = \P -> {s : Str ; p : P} ;
  sp1 : (A : Type) -> Str -> A -> SP1 A = \_,s,a -> {s = s ; p = a} ;

  nonExist : Str = variants {} ;

  optStr : Str -> Str = \s -> variants {s ; []} ;

  constTable : (A,B : Type) -> B -> A => B = \_,_,b -> table {_ => b} ;
  constStr   : (A : Type) -> Str -> A => Str = \A,b -> constTable A Str b ;

  infixSS   : Str  -> SS -> SS -> SS = \f,x,y -> ss (x.s ++ f ++ y.s) ;
  prefixSS  : Str        -> SS -> SS = \f,x   -> ss (f ++ x.s) ;
  postfixSS : Str        -> SS -> SS = \f,x   -> ss (x.s ++ f) ;
  embedSS   : Str -> Str -> SS -> SS = \f,g,x -> ss (f ++ x.s ++ g) ;

-- discontinuous
  SD2 = {s1,s2 : Str} ;
  sd2 : (_,_ : Str) -> SD2 = \x,y -> {s1 = x ; s2 = y} ;

-- parentheses
  paren : Str -> Str = \s -> "(" ++ s ++ ")" ;
  parenss : SS -> SS = \s -> ss (paren s.s) ;

-- free order between two strings
  bothWays : Str -> Str -> Str = \x,y -> variants {x ++ y ; y ++ x} ;

-- parametric order between two strings
  preOrPost : Bool -> Str -> Str -> Str = \pr,x,y -> 
    if_then_else Str pr (x ++ y) (y ++ x) ;

-- Booleans

param Bool = True | False ;

oper
  if_then_else : (A : Type) -> Bool -> A -> A -> A = \_,c,d,e -> 
    case c of {
      True => d ;
      False => e
     } ;

  andB : (_,_ : Bool) -> Bool = \a,b -> if_then_else Bool a b False ;
  orB  : (_,_ : Bool) -> Bool = \a,b -> if_then_else Bool a True b ;
  notB : Bool         -> Bool = \a   -> if_then_else Bool a False True ;


-- zero, one, two, or more (elements in a list etc)

param
  ENumber = E0 | E1 | E2 | Emore ;

oper
  eNext : ENumber -> ENumber = \e -> case e of {
    E0 => E1 ; E1 => E2 ; _ => Emore} ;




--------------------------


-- precedence parameters and operations adapted to the OCL grammars
-- which use {s0, s1 : Str} as the starting point for the linearization 
-- type of expressions.

-- include prelude.gf;

param Prec = p11 | p10 | p9 | p8 | p7 | p6 | p5 | p4 | p3 | p2 | p1;

--
-- OCL precedences:
--
oper atpreP : Prec = p10;
    dotP : Prec = p9;   -- .  ->
    unP : Prec = p8;   -- unary 'not' and '-'
    multP : Prec = p7;
    addP : Prec = p6;
    ifP : Prec = p5; -- if then else endif
    gtP : Prec = p4; -- < > <= >=
    eqP : Prec = p3; -- = <>
    andP : Prec = p2; -- and or xor
    impP : Prec = p1; -- implies

--
-- basic precedence definitions
--

oper nextPrec : Prec => Prec = 
	table {p1 => p2; p2 => p3; p3 => p4; p4 => p5; p5 => p6;
		p6 => p7; p7 => p8; p8 => p9; p9 => p10; p10 => p11; p11 => p11};  -- note p11 => p11

oper mkParenth : Str -> Str  = \str -> "(" ++ str ++ ")";
  
oper mkPrec : Str -> Prec => Prec => Str = \s ->
	let {ps:Str = mkParenth s} in
	table {
	p11 => table {
		_ => s};   -- highest priority, no par.
	p10 => table {
		p11 => ps;
		_ => s};
	p9 => table {
		p11 => ps;
		p10 => ps;
		_ => s};
	p8 => table {
		p11 => ps;
		p10 => ps;
		p9 => ps;
		_ => s};
	p7 => table {
		p11 => ps;
		p10 => ps;
		p9 => ps;
		p8 => ps;
		_ => s};
	p6 => table {
		p11 => ps;
		p10 => ps;
		p9 => ps;
		p8 => ps;
		p7 => ps;
		_ => s}; 
	p5 => table {
		p5 => s;
		p4 => s;
		p3 => s;
		p2 => s;
		p1 => s;
		_ => ps};
	p4 => table {
		p4 => s;
		p3 => s;
		p2 => s;
		p1 => s;
		_ => ps};		 
	p3 => table {
		p3 => s;
		p2 => s;
		p1 => s;
		_ => ps};
	p2 => table {
		p2 => s;
		p1 => s;
		_ => ps};
	p1 => table {
		p1 => s;    -- lowest, without parens only in this case
		_ => ps}
	};



--
-- precedence via tables:
-- 

lintype PrecTerm = {s : Prec => Str};

oper mkConst' : Str -> PrecTerm =
	\c -> {s = mkPrec c ! p11};
    
    mkInfix' : Prec -> Str -> PrecTerm -> PrecTerm -> PrecTerm =
	\p, f, x, y -> 
	let {pp : Prec = nextPrec ! p} 
	in {s = mkPrec (x.s!pp ++ f ++ y.s!pp) ! p};

    mkInfixL' : Prec -> Str -> PrecTerm -> PrecTerm -> PrecTerm =
	\p, f, x, y ->
	{s = mkPrec (x.s!p ++ f ++ y.s!(nextPrec ! p)) ! p};
    mkInfixR' : Prec -> Str -> PrecTerm -> PrecTerm -> PrecTerm =
	\p, f, x, y ->
	{s = mkPrec (x.s!(nextPrec ! p) ++ f ++ y.s!p) ! p};






--
-- for precedence as inherent feature
--

lintype PrecInh = {s0, s1 : Str; p : Prec; isAtpre : Bool};

oper usePrec : PrecInh -> Prec -> Str =
	\x, p -> case x.isAtpre of {
		False => mkPrec x.s0 ! x.p ! p;
		True => mkPrec x.s1 ! x.p ! p
	};


oper noParen : PrecInh -> Str = \x -> usePrec x p1;


oper mkConst : Str -> Str -> PrecInh =
	\c, cAtpre -> {s0 = c; s1 = cAtpre; p = p11; isAtpre = False};

oper mkInfix : Prec -> Str -> Str -> PrecInh -> PrecInh -> PrecInh =
	\p, f, fAtpre, x, y -> {
		s0 = usePrec x (nextPrec ! p) ++ f ++ usePrec y (nextPrec ! p);
		s1 = usePrec x (nextPrec ! p) ++ fAtpre ++ usePrec y (nextPrec ! p);
		p = p;
		isAtpre = False
	};

oper mkInfixL : Prec -> Str -> Str -> PrecInh -> PrecInh -> PrecInh =
	\p, f, fAtpre, x, y -> {
		s0 = usePrec x p ++ f ++ usePrec y (nextPrec ! p);
		s1 = usePrec x p ++ fAtpre ++ usePrec y (nextPrec ! p);
		p = p;
		isAtpre = False
	};

oper mkInfixR : Prec -> Str -> Str -> PrecInh -> PrecInh -> PrecInh =
	\p, f, fAtpre, x, y -> {
		s0 = usePrec x (nextPrec ! p) ++ f ++ usePrec y p;
		s1 = usePrec x (nextPrec ! p) ++ fAtpre ++ usePrec y p;
		p = p;
		isAtpre = False
	};

oper mkPostfix : Prec -> Str -> Str -> PrecInh -> PrecInh =
	\p,f, fAtpre, x -> {
		s0 = usePrec x (nextPrec ! p) ++ f;
		s1 = usePrec x (nextPrec ! p) ++ fAtpre;
		p = p;
		isAtpre = False
	};
	
oper mkPrefix : Prec -> Str -> Str -> PrecInh -> PrecInh =
	\p,f, fAtpre, x -> {
		s0 = f ++ usePrec x (nextPrec ! p);
		s1 = fAtpre ++ usePrec x (nextPrec ! p);
		p = p;
		isAtpre = False
	};
	



--------------------------------------


--
-- classes
--
lincat Class = {s : Str};
oper stdClass : Str -> {s:Str} = ss;

--
-- Subtyping 
--
lin coerce sub super proof obj = obj;
pattern	subTypeRefl c = ["by reflexitivity,"] ++ c ++ ["is a subtype of itself"];
	subTypeTrans a b c proofab proofbc = ["by transitivity,"] ++ a++ ["is a subtype of"] ++
		c ++ ["since"] ++ b ++ ["is a subtype of"] ++ c ++ ["and"] ++ a ++["is a subtype of"] ++ b;

printname coerce = "coercion";

--
-- constraint
-- 
lintype ConstraintL = {s: Str};
lincat Constraint = ConstraintL;
lincat ConstraintHideContext = ConstraintL;

oper contextOp00 : Str -> Str -> Str -> ConstraintL =
		\class, op, body -> ss ("context" ++ class ++ "::" ++ op ++ "()"  ++ body);
	contextOp01 : (class, op, ret, body : Str) -> ConstraintL =
		\class, op, ret, body -> ss ("context" ++ class ++ "::" ++ op ++ "()" ++ ":" ++ ret ++ body);
	contextOp10 : (class, op, args, body : Str) -> ConstraintL = 
		\class, op, args, body -> ss ("context" ++ class ++ "::" ++ op ++ "(" ++ args ++ ")" 
			++ body);
	contextOp11 : Str -> Str -> Str -> Str -> Str -> ConstraintL = 
		\class, op, args, ret, body -> ss ("context" ++ class ++ "::" ++ op ++ "(" ++ args ++ ")" 
			++ ":" ++ ret  ++ body);

	contextClass : Str -> Str -> ConstraintL = \class, body -> 
		ss ("context" ++ class ++ body);


--
-- pre- and postconditions
--
lin
	emptyC = ss "";
	preC cond list = ss ("pre:" ++ inst2str cond ++ list.s);
	postC cond list = ss ("post:" ++ inst2str cond ++ list.s);
	invC cond list = ss ("inv:" ++ inst2str cond ++ list.s);
	
	invHC cond = ss (inst2str cond);
--
-- basics for the type Instance
--
-- discontinous const s0, s1: without and with "@pre", respectively.
-- isAtpre says which one of s0 and s1 is relevant.
lintype Inst = {s0, s1 : Str; p : Prec; isAtpre : Bool};  
lincat Instance = Inst;
oper inst2str : Inst -> Str = \i -> case i.isAtpre of {
	True => i.s1;
	False => i.s0
	};

-- should we have ["@ pre"] instead of "@pre"?
lin atPre _ i = {s0 = i.s0; s1 = i.s1; p = i.p; isAtpre = True};
printname atPre = "@pre";

oper const : Str -> Inst = \f -> mkConst f (f ++ "@pre");
	 postfix : Prec -> Str -> Inst -> Inst = \p, f, a -> 
	 	mkPostfix p f (f ++ "@pre") a ;
	 prefix : Prec -> Str -> Inst -> Inst = \p, f, a->
	 	mkPrefix p f (f ++ "@pre") a;
	 infix : Prec -> Str -> Inst -> Inst -> Inst = \p, f, a, b ->
	 	mkInfix p f (f ++ "@pre") a b;
	 infixL : Prec -> Str -> Inst -> Inst -> Inst = \p, f, a, b ->
	 	mkInfixL p f (f ++ "@pre") a b;
 	 infixR : Prec -> Str -> Inst -> Inst -> Inst = \p, f, a, b ->
 	 	mkInfixR p f (f ++ "@pre") a b;


{--- deprecated operations related to properties, use the ones just below instead
oper dot : Inst -> Inst -> Inst = \a,b -> infixL dotP "." a b;
	 dotAttr : Inst -> Str -> Inst = \a,s -> infixL dotP "." a (const s);
	 dotAssoc : Inst -> Str -> Inst = 
		\obj,assoc -> infixL dotP "." obj (const assoc);
	dotAssocQ : Inst -> Str -> Inst -> Inst = \obj,assoc,qual ->
		infixL dotP "." obj (const (assoc ++ "[" ++ noParen qual ++ "]"));
	dotAssocClass : Inst -> Str -> Str -> Inst =
		\obj,assoc,role -> infixL dotP "." obj 
			(const (variants {assoc; assoc ++ "[" ++ role ++ "]"}));
	dotAssocClassRec : Inst -> Str -> Str -> Inst =
		\obj,assoc,role -> infixL dotP "." obj (const (assoc ++ "[" ++ role ++ "]"));
	 arrow1 : Inst -> Inst -> Inst = \a,b -> infixL dotP "->" a b;
	 arrowAttr : Inst -> Str -> Inst = \a,s -> infixL dotP "->" a (mkConst s);
	 appP : Prec = p11; 
	 app0 : Str -> Inst = \operation -> {s0 = operation; s1 = "()"; p = appP};
	 app1 : Str -> Inst -> Inst = \operation, x -> 
		{s0 = operation; s1 =  "(" ++ noParen x ++ ")"; p = appP};
	enumElem : Str -> Str -> Inst = \c,e -> const (c ++ "::" ++ e);
-}

-- new and fabulous operations related to properties:

-- build list of arguments/parameters to queries:	
lintype  ArgList = {s : Str; empty : Bool};
oper argNil : ArgList = {s = ""; empty = True};
	argCons : Inst -> ArgList -> ArgList = 
		\x, xs -> {s = case xs.empty of {
			True => noParen x;
			False => xs.s ++ "," ++ noParen x
		}; empty = False};
	dotOper : Inst -> Str -> Str -> Inst = \obj, query, args -> { 
		s0 = (infixL dotP "." obj (const (query ++ "(" ++ args ++")"))).s0;
		s1 = (infixL dotP "." obj (const (query ++ "@pre" ++ "(" ++ args ++")"))).s0;
		isAtpre = False;
		p = dotP};
	arrowOper : Inst -> Str -> Str -> Inst = \obj, query, args -> { 
		s0 = (infixL dotP "->" obj (const (query ++ "(" ++ args ++")"))).s0;
		s1 = (infixL dotP "->" obj (const (query ++ "@pre" ++ "(" ++ args ++")"))).s0;
		isAtpre = False;
		p = dotP};
	dotAssoc : Inst -> Str -> Inst = \obj, assoc -> { 
		s0 = (infixL dotP "." obj (const assoc )).s0;
		s1 = (infixL dotP "." obj (const (assoc ++ "@pre"))).s0;
		isAtpre = False;
		p = dotP};
	dotOper0 : Inst -> Str -> Inst = \obj, query -> {
		s0 = (infixL dotP "." obj (const (query ++ "()"))).s;
		s1 = (infixL dotP "." obj (const (query ++ "@pre" ++ "()"))).s;
		isAtpre = False;
		p = dotP
	};
	dotOper1 : Inst -> Str -> Inst -> Inst = \obj, query, arg -> {
		s0 = (infixL dotP "." obj (const (query ++ "(" ++ noParen arg ++ ")"))).s;
		s1 = (infixL dotP "." obj (const (query ++ "@pre" ++ "(" ++ noParen arg ++ ")"))).s;
		isAtpre = False;
		p = dotP};
	arrowOper0 : Inst -> Str -> Inst = \obj, query -> {
		s0 = (infixL dotP "->" obj (const (query ++ "()"))).s;
		s1 = (infixL dotP "->" obj (const (query ++ "@pre" ++ "()"))).s;
		isAtpre = False;
		p = dotP
	};
	arrowOper1 : Inst -> Str -> Inst -> Inst = \obj, query, arg -> {
		s0 = (infixL dotP "->" obj (const (query ++ "(" ++ noParen arg ++ ")"))).s;
		s1 = (infixL dotP "->" obj (const (query ++ "@pre" ++ "(" ++ noParen arg ++ ")"))).s;
		isAtpre = False;
		p = dotP};		
	dotAttr : Inst -> Str -> Inst = \a,s -> {
		s0 = (infixL dotP "." a (const s)).s0;
		s1 = (infixL dotP "." a (const (s ++ "@pre"))).s0;
		isAtpre = False;
		p = dotP};
	arrowAttr : Inst -> Str -> Inst = \a,s -> {
		s0 = (infixL dotP "->" a (const s)).s0;
		s1 = (infixL dotP "->" a (const (s ++ "@pre"))).s0;
		isAtpre = False;
		p = dotP};

	mkQuery : Inst -> Str -> ArgList -> Inst = \obj, query, args -> 
		dotOper obj query args.s;

--
-- Sent (should be the same as Instance)
--
lincat Sent = Inst;
lin atPreSent i = {s0 = i.s0; s1 = i.s1; p = i.p; isAtpre = True};

lin sent2bool s = s;
	bool2sent b = b;

printname sent2bool = "use a sentence as a boolean";
	  bool2sent = "use a boolean as a sentence";

--
-- self, result 
--
lin
	self _ _ = const "self";
	result _ _ = const "result";
	

--
-- let-expressions
--
lincat LetExp = {s0, s1 : Str}; -- s0 for the definitions, s1 for the body
	LetExpSent = {s0, s1 : Str};
	
lin letSingleAttr c _ defn body = {
		s0 = "let" ++ body.v0 ++ ":" ++ c.s ++ "=" ++ inst2str defn;
		s1 = body.s};
	letConsAttr c _  defn body = {
		s0 = "let" ++ body.v0 ++ ":" ++ c.s ++ "=" ++ inst2str defn ++ body.s0;
		s1 = body.s1};
	letSingleOp1arg c1 c2 c3 defn body = {
		s0 = "let" ++ body.v0 ++ "(" ++ defn.v0 ++ ":" ++ c1.s ++ ")" ++
			":" ++ c2.s ++ "=" ++ (inst2str defn);
		s1 = body.s};
		
	letExp2Inst _ le = const (le.s0 ++ "in" ++ le.s1);

	letSingleAttrS c defn body = {
		s0 = "let" ++ body.v0 ++ ":" ++ c.s ++ "=" ++ inst2str defn;
		s1 = body.s};
	letConsAttrS c defn body = {
		s0 = "let" ++ body.v0 ++ ":" ++ c.s ++ "=" ++ inst2str defn ++ body.s0;
		s1 = body.s1};
		
	letExp2Sent le = const (le.s0 ++ "in" ++ le.s1);



-------------


--
-- OclType
--
lin OclType = stdClass "OclType";

    class2oclType c = const c.s;

    otName t = dotOper0 t "name";
    otAttributes t = dotOper0 t "attributes";
    otOperations t = dotOper0 t "operations";
    otSupertypes t = dotOper0 t "supertypes";
    otAllSupertypes t = dotOper0 t "allSupertypes";
    otAllInstancesStrict _ t = dotOper0 t "allInstances";


--
-- OclAny 
--
lin OclAny = stdClass "OclAny";
    conforms2OclAny c = ss (c.s ++ ["conforms to OclAny"]);
    
    anyEq a b =  infix eqP "=" a b;
    anyNeq a b =  infix eqP "<>" a b;
    anyOclIsKindOf a t = dotOper1 a "oclIsKindOf" t;
    anyOclIsTypeOf a t = dotOper1 a "oclIsTypeOf" t;
    anyOclAsTypeStrict a c t = dotOper1 a "oclAsType" t;
    anyOclInState a s = dotOper1 a "oclIsInState" s;
    anyOclIsNew a = dotOper0 a "oclIsNew";


--
-- OclState 
--
lin OclState = stdClass "OclState";


--
-- OclExpression
--
lin OclExpression = stdClass "OclExpression";

    inst2OclExpr _ i = i; -- oh yes, circular
    oeEvaluationType e = dotOper0 e "evaluationType";


--
-- Integer
--
lin Integer = stdClass "Integer";
lin	
	intEq a b = infix eqP "=" a b;
	intAdd a b = infix addP "+" a b;
	intSub a b = infixL addP "-" a b;
	intUnaryMinus a = prefix unP "-" a;
	intTimes a b = infix multP "*" a b;
	intDiv2Real a b = infixL multP "/" a b;
	intAbs a = dotOper0 a "abs";
	intDiv2Int a b = dotOper1 a "div" b;
	intMod a b = dotOper1 a "mod" b;
	intMax a b = dotOper1 a "max" b;
	intMin a b = dotOper1 a "min" b;
	intLT a b = infix gtP "<" a b;
	intGT a b = infix gtP ">" a b;
	intLTE a b = infix gtP "<=" a b;
	intGTE a b = infix gtP ">=" a b;

lin	Zero = const "0";
	One = const "1";
	Two = const "2";
	int i = const i.s;

--
-- Real
--
lin Real = stdClass "Real";
lin
	realEq a b = infix eqP "=" a b;
	realNeq a b = infix eqP "<>" a b;
	realAdd a b = infix addP "+" a b;
	realSub a b = infixL addP "-" a b;
	realUnaryMinus a = prefix unP "-" a;
	realTimes a b = infix multP "*" a b;
	realDiv a b = infixL multP "/" a b;
	realAbs a = dotOper0 a "abs";
	realFloor a = dotOper0 a "floor";
	realRound a = dotOper0 a "round";
	realMax a b = dotOper1 a "max" b;
	realMin a b = dotOper1 a "min" b;
	realLT a b = infix gtP "<" a b;
	realGT a b = infix gtP ">" a b;
	realLTE a b = infix gtP "<=" a b;
	realGTE a b = infix gtP ">=" a b;


--
-- Bool
--
lin BoolC = stdClass "Bool";
	true = const "true";
	false = const "false";
	notBool a = prefix unP "not" a;
	boolEq a b = infix eqP "=" a b;
	orBool a b = infix andP "or" a b;
	andBool a b = infix andP "and" a b;
	xorBool a b = infix andP "xor" a b;
	impliesBool a b = infixR impP "implies" a b; 
--
-- Sent
--
lin sentEq a b = infix eqP "=" a b;
	notS a = prefix unP "not" a;
	orS a b = infix andP "or" a b;
	andS a b = infix andP "and" a b;
	xorS a b = infix andP "xor" a b;
	impliesS a b = infixR impP "implies" a b; 

lin ifThenElse _ cond t e = {
		s0 = "if" ++ noParen cond ++ "then" ++ noParen t ++ "else" ++ noParen e ++ "endif"; 
		s1 = "if" ++ noParen cond ++ "then" ++ noParen t ++ "else" ++ noParen e ++ ["endif @pre"]; -- erroneous 
		p = ifP;
		isAtpre = False
	};
	
	ifThenElseS cond t e = {
		s0 = "if" ++ noParen cond ++ "then" ++ noParen t ++ "else" ++ noParen e ++ "endif"; 
		s1 = "if" ++ noParen cond ++ "then" ++ noParen t ++ "else" ++ noParen e ++ ["endif @pre"]; -- erroneous 
		p = ifP;
		isAtpre = False
	};
	
-- 
-- String
--
lin StringC = stdClass "String";
	stringEq a b = infix eqP "=" a b;
	stringLiteral lit = const (lit.s);
	stringSize s = dotOper0 s "size";
	stringConcat a b = dotOper1 a "concat" b;
	stringToUpper s = dotOper0 s "toUpper";
	stringToLower s = dotOper0 s "toLower";
	stringSubstring s a b = dotOper s "substring" (noParen a ++ "," ++ noParen b);

--
-- Iterators (N.B.: not used for the OCL "iterate" construction!)
--
lincat InstIter = {s1, s2, s3 : Str};
{---- with variants:
oper wrapIter : {s1, s2, s3 :Str} -> Str = \i -> (variants {
		i.s1 ++ i.s2 ++ "|" ++ i.s3;  -- with type info (i.s2)
		i.s1 ++ "|" ++ i.s3});     -- without type info
----}
-- without variants (just one of them):
oper wrapIter : {s1, s2, s3 :Str} -> Str = \i -> (
		i.s1 ++ i.s2 ++ "|" ++ i.s3);  -- with type info (i.s2)

		
lin	instIterSingle varC _ body = {s1 = body.v; s2 = ":" ++ varC.s; 
		s3 = inst2str {s0=body.s0; s1=body.s1; p=body.p; isAtpre=body.isAtpre}};
	instIterMany varC _ body = { s1 = body.v ++ "," ++ body.s1; s2 = ":" ++ varC.s;
		s3 = body.s3};
		
lincat SentIter = {s1, s2, s3 : Str};
lin sentIterSingle varC body = {s1 = body.v; s2 = ":" ++ varC.s; 
		s3 = inst2str {s0=body.s0; s1=body.s1; p=body.p; isAtpre=body.isAtpre}};
	sentIterMany varC body = { s1 = body.v ++ "," ++ body.s1; s2 = ":" ++ varC.s;
		s3 = body.s3};
		
--
-- collections
--
lin
	Collection c = stdClass ("Collection" ++ c.s);
	size _ coll = arrowOper0 coll "size";
	includes _ _ coll a = arrowOper1 coll "includes" a; 
	excludes _ _ coll a = arrowOper1 coll "excludes" a;
	count _ coll a = arrowOper1 coll "count" a;
	includesAll _ coll1 coll2 = arrowOper1 coll1 "includesAll" coll2;
	excludesAll _ coll1 coll2 = arrowOper1 coll1 "excludesAll" coll2;
	isEmpty _ coll = arrowOper coll "isEmpty" "";
	notEmpty _ coll = arrowOper coll "notEmpty" "";
	sum _ coll = arrowOper coll "sum" "";
	exists c coll i = arrowOper coll "exists" (wrapIter i);
		-- arrow1 coll (app1 "exists" (wrapIter i));
	forAll c coll i = arrowOper coll "forAll" (wrapIter i);
	isUnique c d coll i = arrowOper coll "isUnique" (wrapIter i);
	sortedBy c d coll i = arrowOper coll "sortedBy" (wrapIter i);
	iterate c a coll init body = arrowOper coll "iterate" ( 
		body.v0 ++ ":" ++ c.s ++ ";" ++    -- this is the "element", same type as the collection
		body.v1 ++ ":" ++ a.s ++ "=" ++    -- this is the "accumulator", same type as the result
		(noParen init) ++ "|" ++
		(noParen {s=body.s; s1=body.s1; p=body.p; isAtpre = False})  -- NB: "noParen body"  does not work with GF!?
		);
	any c coll i = arrowOper coll "any" (wrapIter i);
	one c coll i = arrowOper coll "one" (wrapIter i);
	
--
-- Set
--
lin	Set c = stdClass ("Set" ++ c.s);
	unionSS c set1 set2 = arrowOper1 set1  "union" set2;
	unionSB c set bag = arrowOper1 set "union" bag;
	setEq c set1 set2 = infix eqP "=" set1 set2;
	intersectionSS c set1 set2 = arrowOper1 set1 "intersection" set2;
	intersectionSB c set bag = arrowOper1 set "intersection" bag;
	subS c set1 set2 = infixL addP "-" set1 set2;
	includingS c set a = arrowOper1 set "including" a;
	excludingS c set a = arrowOper1 set "excluding" a;
	symmetricDiff c set1 set2 = arrowOper1 set1 "symmetricDifference" set2;
	selectSet c set i = arrowOper set "select" (wrapIter i);
	rejectSet c set i = arrowOper set "reject" (wrapIter i);
	collectSet c d set i = arrowOper set "exists" (wrapIter i); 
	countSet c set a = arrowOper1 set "count" a;
	setAsSequence c set = arrowOper set "asSequence" "";
	setAsBag c set = arrowOper set "asBag" "";

--
-- Bag
--
lin	Bag c = stdClass ("Bag" ++ c.s);
	bagEq _ a b = infix eqP "=" a b;
	unionBB _ a b = arrowOper1 a "union" b;
	unionBS _ a b = arrowOper1 a "union" b;
	intersectionBB _ a b = arrowOper1 a "intersection" b;
	intersectionBS _ a b = arrowOper1 a "intersection" b;
	includingB _ bag e = arrowOper1 bag "including" e;
	excludingB _ bag e = arrowOper1 bag "excluding" e;
	selectBag _ bag i = arrowOper bag "select" (wrapIter i);
	rejectBag _ bag i = arrowOper bag "reject" (wrapIter i);
	collectBag _ _  bag i = arrowOper bag "collect" (wrapIter i);
	countBag _ bag e = arrowOper1 bag "count" e;
	bagAsSequence _ bag = arrowOper0 bag "asSequence";
	bagAsSet _ bag = arrowOper0 bag "asSet";
	
		


--
-- Sequence
--
lin	Sequence c = stdClass ("Sequence" ++ c.s);
	sequenceEq c a b = infix eqP "=" a b;
	unionSeq _ a b = arrowOper1 a "union" b;
	append _ a b = arrowOper1 a "append" b;
	prepend _ a b = arrowOper1 a "prepend" b;
	subSequence _ a start end = arrowOper a "subSequence" (noParen start ++ "," ++ noParen end);
	at _ a i = arrowOper1 a "at" i;
	first _ a = arrowOper0 a "first";
	last _ a = arrowOper0 a "last";
	includingSeq _ seq e = arrowOper1 seq "including" e;
	excludingSeq _ seq e = arrowOper1 seq "excluding" e;
	selectSeq _ seq i = arrowOper seq "select" (wrapIter i);
	rejectSeq _ seq i = arrowOper seq "reject" (wrapIter i);
	collectSeq _ _  seq i = arrowOper seq "collect" (wrapIter i);
	seqAsBag _ a = arrowOper0 a "asBag";
	seqAsSet _ a = arrowOper0 a "asSet";
