include 
	prelude/prelude.gf;
	precedence.gf;

--
-- classes
--
lincat Class = {s : Str};
oper stdClass : Str -> {s:Str} = ss;

--
-- Subtyping 
--
lin coerce sub super proof obj = obj;
pattern	subTypeRefl c = ["by reflexitivity,"] ++ c ++ ["is a subtype of itself"];
	subTypeTrans a b c proofab proofbc = ["by transitivity,"] ++ a++ ["is a subtype of"] ++
		c ++ ["since"] ++ b ++ ["is a subtype of"] ++ c ++ ["and"] ++ a ++["is a subtype of"] ++ b;

printname coerce = "coercion";

--
-- constraint
-- 
lintype ConstraintL = {s: Str};
lincat Constraint = ConstraintL;
lincat ConstraintHideContext = ConstraintL;

oper contextOp00 : Str -> Str -> Str -> ConstraintL =
		\class, op, body -> ss ("context" ++ class ++ "::" ++ op ++ "()"  ++ body);
	contextOp01 : (class, op, ret, body : Str) -> ConstraintL =
		\class, op, ret, body -> ss ("context" ++ class ++ "::" ++ op ++ "()" ++ ":" ++ ret ++ body);
	contextOp10 : (class, op, args, body : Str) -> ConstraintL = 
		\class, op, args, body -> ss ("context" ++ class ++ "::" ++ op ++ "(" ++ args ++ ")" 
			++ body);
	contextOp11 : Str -> Str -> Str -> Str -> Str -> ConstraintL = 
		\class, op, args, ret, body -> ss ("context" ++ class ++ "::" ++ op ++ "(" ++ args ++ ")" 
			++ ":" ++ ret  ++ body);

	contextClass : Str -> Str -> ConstraintL = \class, body -> 
		ss ("context" ++ class ++ body);


--
-- pre- and postconditions
--
lin
	emptyC = ss "";
	preC cond list = ss ("pre:" ++ inst2str cond ++ list.s);
	postC cond list = ss ("post:" ++ inst2str cond ++ list.s);
	invC cond list = ss ("inv:" ++ inst2str cond ++ list.s);
	
	invHC cond = ss (inst2str cond);
--
-- basics for the type Instance
--
-- discontinous const s0, s1: without and with "@pre", respectively.
-- isAtpre says which one of s0 and s1 is relevant.
lintype Inst = {s0, s1 : Str; p : Prec; isAtpre : Bool};  
lincat Instance = Inst;
oper inst2str : Inst -> Str = \i -> case i.isAtpre of {
	True => i.s1;
	False => i.s0
	};

-- should we have ["@ pre"] instead of "@pre"?
lin atPre _ i = {s0 = i.s0; s1 = i.s1; p = i.p; isAtpre = True};
printname atPre = "@pre";

oper const : Str -> Inst = \f -> mkConst f (f ++ "@pre");
	 postfix : Prec -> Str -> Inst -> Inst = \p, f, a -> 
	 	mkPostfix p f (f ++ "@pre") a ;
	 prefix : Prec -> Str -> Inst -> Inst = \p, f, a->
	 	mkPrefix p f (f ++ "@pre") a;
	 infix : Prec -> Str -> Inst -> Inst -> Inst = \p, f, a, b ->
	 	mkInfix p f (f ++ "@pre") a b;
	 infixL : Prec -> Str -> Inst -> Inst -> Inst = \p, f, a, b ->
	 	mkInfixL p f (f ++ "@pre") a b;
 	 infixR : Prec -> Str -> Inst -> Inst -> Inst = \p, f, a, b ->
 	 	mkInfixR p f (f ++ "@pre") a b;


{--- deprecated operations related to properties, use the ones just below instead
oper dot : Inst -> Inst -> Inst = \a,b -> infixL dotP "." a b;
	 dotAttr : Inst -> Str -> Inst = \a,s -> infixL dotP "." a (const s);
	 dotAssoc : Inst -> Str -> Inst = 
		\obj,assoc -> infixL dotP "." obj (const assoc);
	dotAssocQ : Inst -> Str -> Inst -> Inst = \obj,assoc,qual ->
		infixL dotP "." obj (const (assoc ++ "[" ++ noParen qual ++ "]"));
	dotAssocClass : Inst -> Str -> Str -> Inst =
		\obj,assoc,role -> infixL dotP "." obj 
			(const (variants {assoc; assoc ++ "[" ++ role ++ "]"}));
	dotAssocClassRec : Inst -> Str -> Str -> Inst =
		\obj,assoc,role -> infixL dotP "." obj (const (assoc ++ "[" ++ role ++ "]"));
	 arrow1 : Inst -> Inst -> Inst = \a,b -> infixL dotP "->" a b;
	 arrowAttr : Inst -> Str -> Inst = \a,s -> infixL dotP "->" a (mkConst s);
	 appP : Prec = p11; 
	 app0 : Str -> Inst = \operation -> {s0 = operation; s1 = "()"; p = appP};
	 app1 : Str -> Inst -> Inst = \operation, x -> 
		{s0 = operation; s1 =  "(" ++ noParen x ++ ")"; p = appP};
	enumElem : Str -> Str -> Inst = \c,e -> const (c ++ "::" ++ e);
-}

-- new and fabulous operations related to properties:

-- build list of arguments/parameters to queries:	
lintype  ArgList = {s : Str; empty : Bool};
oper argNil : ArgList = {s = ""; empty = True};
	argCons : Inst -> ArgList -> ArgList = 
		\x, xs -> {s = case xs.empty of {
			True => noParen x;
			False => xs.s ++ "," ++ noParen x
		}; empty = False};
	dotOper : Inst -> Str -> Str -> Inst = \obj, query, args -> { 
		s0 = (infixL dotP "." obj (const (query ++ "(" ++ args ++")"))).s0;
		s1 = (infixL dotP "." obj (const (query ++ "@pre" ++ "(" ++ args ++")"))).s0;
		isAtpre = False;
		p = dotP};
	arrowOper : Inst -> Str -> Str -> Inst = \obj, query, args -> { 
		s0 = (infixL dotP "->" obj (const (query ++ "(" ++ args ++")"))).s0;
		s1 = (infixL dotP "->" obj (const (query ++ "@pre" ++ "(" ++ args ++")"))).s0;
		isAtpre = False;
		p = dotP};
	dotAssoc : Inst -> Str -> Inst = \obj, assoc -> { 
		s0 = (infixL dotP "." obj (const assoc )).s0;
		s1 = (infixL dotP "." obj (const (assoc ++ "@pre"))).s0;
		isAtpre = False;
		p = dotP};
	dotOper0 : Inst -> Str -> Inst = \obj, query -> {
		s0 = (infixL dotP "." obj (const (query ++ "()"))).s;
		s1 = (infixL dotP "." obj (const (query ++ "@pre" ++ "()"))).s;
		isAtpre = False;
		p = dotP
	};
	dotOper1 : Inst -> Str -> Inst -> Inst = \obj, query, arg -> {
		s0 = (infixL dotP "." obj (const (query ++ "(" ++ noParen arg ++ ")"))).s;
		s1 = (infixL dotP "." obj (const (query ++ "@pre" ++ "(" ++ noParen arg ++ ")"))).s;
		isAtpre = False;
		p = dotP};
	arrowOper0 : Inst -> Str -> Inst = \obj, query -> {
		s0 = (infixL dotP "->" obj (const (query ++ "()"))).s;
		s1 = (infixL dotP "->" obj (const (query ++ "@pre" ++ "()"))).s;
		isAtpre = False;
		p = dotP
	};
	arrowOper1 : Inst -> Str -> Inst -> Inst = \obj, query, arg -> {
		s0 = (infixL dotP "->" obj (const (query ++ "(" ++ noParen arg ++ ")"))).s;
		s1 = (infixL dotP "->" obj (const (query ++ "@pre" ++ "(" ++ noParen arg ++ ")"))).s;
		isAtpre = False;
		p = dotP};		
	dotAttr : Inst -> Str -> Inst = \a,s -> {
		s0 = (infixL dotP "." a (const s)).s0;
		s1 = (infixL dotP "." a (const (s ++ "@pre"))).s0;
		isAtpre = False;
		p = dotP};
	arrowAttr : Inst -> Str -> Inst = \a,s -> {
		s0 = (infixL dotP "->" a (const s)).s0;
		s1 = (infixL dotP "->" a (const (s ++ "@pre"))).s0;
		isAtpre = False;
		p = dotP};

	mkQuery : Inst -> Str -> ArgList -> Inst = \obj, query, args -> 
		dotOper obj query args.s;

--
-- Sent (should be the same as Instance)
--
lincat Sent = Inst;
lin atPreSent i = {s0 = i.s0; s1 = i.s1; p = i.p; isAtpre = True};

lin sent2bool s = s;
	bool2sent b = b;

printname sent2bool = "use a sentence as a boolean";
	  bool2sent = "use a boolean as a sentence";

--
-- self, result 
--
lin
	self _ _ = const "self";
	result _ _ = const "result";
	

--
-- let-expressions
--
lincat LetExp = {s0, s1 : Str}; -- s0 for the definitions, s1 for the body
	LetExpSent = {s0, s1 : Str};
	
lin letSingleAttr c _ defn body = {
		s0 = "let" ++ body.v0 ++ ":" ++ c.s ++ "=" ++ inst2str defn;
		s1 = body.s};
	letConsAttr c _  defn body = {
		s0 = "let" ++ body.v0 ++ ":" ++ c.s ++ "=" ++ inst2str defn ++ body.s0;
		s1 = body.s1};
	letSingleOp1arg c1 c2 c3 defn body = {
		s0 = "let" ++ body.v0 ++ "(" ++ defn.v0 ++ ":" ++ c1.s ++ ")" ++
			":" ++ c2.s ++ "=" ++ (inst2str defn);
		s1 = body.s};
		
	letExp2Inst _ le = const (le.s0 ++ "in" ++ le.s1);

	letSingleAttrS c defn body = {
		s0 = "let" ++ body.v0 ++ ":" ++ c.s ++ "=" ++ inst2str defn;
		s1 = body.s};
	letConsAttrS c defn body = {
		s0 = "let" ++ body.v0 ++ ":" ++ c.s ++ "=" ++ inst2str defn ++ body.s0;
		s1 = body.s1};
		
	letExp2Sent le = const (le.s0 ++ "in" ++ le.s1);

